indexing

	description:
		"Can be an entire MIME message, or a part of a multipart message."

	author: "Berend de Boer"
	date: "$Date: 2003/05/08 $"
	revision: "$Revision: #2 $"


class

	EPX_MIME_PART


inherit

	STDC_BASE

	EPX_MIME_FIELD_NAMES
		export
			{NONE} all
		end

	EPX_MIME_PARAMETER_NAMES
		export
			{NONE} all
		end


creation

	make_empty


feature -- Initialization

	make_empty is
			-- Make an empty MIME part. Useful during parsing to
			-- construct the message gradually.
		do
			create header.make_default
		end


feature -- Output

	append_to_string (s: STRING) is
			-- Stream contents of MIME structure to `s'.
			-- Adds the Content-Length field with the proper length, if
			-- this field does not exist. If there is a Content-Length
			-- field, it will assume that its size is correct.
			-- Line lengths should remain within the limit set by RFC
			-- 822, i.e. no more than 998 characters, and preferably no
			-- more than 78 (this length excludes the CRLF).
		require
			s_not_void: s /= Void
			multipart_implies_have_boundary:
				body /= Void and then body.is_multipart implies
					(header.content_type /= Void implies
						(header.content_type.boundary /= Void and then
						 not header.content_type.boundary.is_empty))
		local
			ct: EPX_MIME_FIELD_CONTENT_TYPE
			body_text: STRING
			cl: EPX_MIME_FIELD_CONTENT_LENGTH
			body_size: INTEGER
			old_s_count: INTEGER
			has_boundary: BOOLEAN
		do
			ct := header.content_type
			if body /= Void then
				create body_text.make_empty
				body.append_to_string (body_text)
				-- If there is a Content-Length field, assume it is correct.
				if not header.has (field_name_content_length) then
					body_size := body_text.count
					if ct /= Void then
						has_boundary := ct.boundary /= Void
						if has_boundary then
							body_size :=
								body_size +
								2 + ct.boundary.count + once_crlf.count +
								2 + ct.boundary.count + 2 + once_crlf.count
						end
					end
					create cl.make (body_size)
					header.add_field (cl)
				end
			end
			header.append_fields_to_string (s)
			s.append_string (once_crlf)
			old_s_count := s.count
			if has_boundary then
				s.append_character ('-')
				s.append_character ('-')
				s.append_string (ct.boundary)
				s.append_string (once_crlf)
			end
			if body_text /= Void then
				s.append_string (body_text)
			end
			if has_boundary then
				s.append_character ('-')
				s.append_character ('-')
				s.append_string (ct.boundary)
				s.append_character ('-')
				s.append_character ('-')
				s.append_string (once_crlf)
			end

				check
					content_length_correct: body_size = s.count - old_s_count
				end
		end

	append_urlencoded_to_string (s: STRING) is
			-- Append fields of this MIME structure to `s', and the body
			-- as x-www-form-urlencoded.
		require
			s_not_void: s /= Void
			multipart_body: body = Void or else body.is_multipart
			has_no_parts_with_multipart_bodies: body = Void or else not body.has_parts_with_multipart_bodies
			every_part_has_a_form_content_disposition_field: body = Void or else body.has_every_part_a_form_content_disposition_field
		local
			multipart_body: EPX_MIME_BODY_MULTIPART
			body_text: STRING
			cl: EPX_MIME_FIELD_CONTENT_LENGTH
		do
			multipart_body ?= body
			if multipart_body /= Void then
				create body_text.make_empty
				multipart_body.append_urlencoded_to_string (body_text)
				-- If there is a Content-Length field, assume it is correct.
				if not header.has (field_name_content_length) then
					create cl.make (body_text.count)
					header.add_field (cl)
				end
			end
			header.append_fields_to_string (s)
			s.append_string (once_crlf)
			if body_text /= Void then
				s.append_string (body_text)
			end
		end


feature -- Access

	header: EPX_MIME_HEADER
			-- Fields for this part.

	body: EPX_MIME_BODY
			-- The body, can be multipart.


feature -- Body creation/removal

	clear_body is
			-- Set `body' to Void.
		do
			body := Void
		ensure
			body_void: body = Void
		end

	create_multipart_body is
			-- Set `body' to a container.
		require
			body_not_set: body = Void
		do
			create {EPX_MIME_BODY_MULTIPART} body.make
		ensure
			body_not_void: body /= Void
			body_is_multipart: body.is_multipart
		end

	create_singlepart_body is
			-- Set `body' to a single part body.
			-- If we find a Content-Disposition field with value
			-- form-data, the data is saved to a temporary file.
		require
			body_not_set: body = Void
		local
			cd: EPX_MIME_FIELD_CONTENT_DISPOSITION
		do
			cd := header.content_disposition
			if cd = Void then
				create {EPX_MIME_BODY_STRING} body.make
			else
				if cd.parameters.has (parameter_name_filename) then
					create {EPX_MIME_BODY_FILE} body.make
				else
					create {EPX_MIME_BODY_STRING} body.make
				end
			end
		ensure
			body_not_void: body /= Void
			body_is_singlepart: not body.is_multipart
		end


feature {NONE} -- Once strings

	once_crlf: STRING is "%R%N"


invariant

	header_not_void: header /= Void


end
