indexing

	description: "Class that lets you safely create an XML document."

	notes: "This is not DOM, but a kind of streaming XML generation."

	known_bugs: "Does not output correct XML for mixed content nodes."

	author: "Berend de Boer"
	date: "$Date: 2004/01/02 $"
	revision: "$Revision: #11 $"


class

	EPX_XML_WRITER


inherit

	STDC_BASE

	EPX_XML_ROUTINES

	KL_IMPORTED_STRING_ROUTINES
		export
			{NONE} all
		end

	UT_CHARACTER_CODES
		export
			{NONE} all
		end


creation

	make,
	make_with_capacity,
	make_fragment,
	make_fragment_with_capacity


feature -- Initialization

	make is
			-- Create an XML document with initial capacity of 1024 characters.
		do
			make_with_capacity (1024)
		ensure
			no_tag_started: not is_tag_started
			as_string_empty: as_string.is_empty
			as_uc_string_empty: as_uc_string.is_empty
			header_written: is_fragment = is_header_written
			not_fragment: not is_fragment
		end

	make_fragment is
			-- Create an XML fragment (document without header) with
			-- initial capacity of 1024 characters.
		do
			make_fragment_with_capacity (1024)
		ensure
			no_tag_started: not is_tag_started
			as_string_empty: as_string.is_empty
			as_uc_string_empty: as_uc_string.is_empty
			header_written: is_fragment = is_header_written
			fragment: is_fragment
		end

	make_with_capacity (a_capacity: INTEGER) is
			-- Create an XML document with initial capacity of
			-- `a_capacity' characters.
		local
			tester: UC_EQUALITY_TESTER
		do
			is_fragment := False
			create my_xml.make (a_capacity)
			create my_xml.make_empty
			create {DS_LINKED_STACK [STRING]} tags.make
			create tester
			tags.set_equality_tester (tester)
			create attributes.make
			attributes.set_equality_tester (tester)
			create values.make (48)
			values.set_key_equality_tester (tester)
			clear
		ensure
			no_tag_started: not is_tag_started
			as_string_empty: as_string.is_empty
			as_uc_string_empty: as_uc_string.is_empty
			header_written: is_fragment = is_header_written
			not_fragment: not is_fragment
		end

	make_fragment_with_capacity (a_capacity: INTEGER) is
			-- Create an XML fragment (document without header) with
			-- initial capacity of `a_capacity' characters.
		do
			make_with_capacity (a_capacity)
			is_fragment := True
			is_header_written := True
		ensure
			no_tag_started: not is_tag_started
			as_string_empty: as_string.is_empty
			as_uc_string_empty: as_uc_string.is_empty
			header_written: is_fragment = is_header_written
			fragment: is_fragment
		end


feature -- Status

	is_a_parent (tag: STRING): BOOLEAN is
			-- Is `tag' the current element, or is it a parent of the
			-- current tag at some point?
		do
			Result := tags.has (tag)
		ensure
			started_implies_parent: is_started (tag) implies Result
		end

	is_element_with_data: BOOLEAN
			-- Has data been added to this element or in case this
			-- element has not yet been written, has data been added to
			-- its parents element?

	is_fragment: BOOLEAN
			-- Is the XML document being created a fragment?

	is_header_written: BOOLEAN
			-- Is the XML header is written or is this a fragment that
			-- does not need a header?

	is_ns_started (a_name_space, a_tag: STRING): BOOLEAN is
			-- Is name_space:tag the current element?
		require
			a_tag_not_void: a_tag /= Void
		local
			s: STRING
		do
			if a_name_space /= Void and then not a_name_space.is_empty then
				s := clone (a_name_space)
				s.append_character (':')
				s.append_string (a_tag)
				Result := is_started (s)
			else
				Result := is_started (a_tag)
			end
		end

	is_started (a_tag: STRING): BOOLEAN is
			-- Is `tag' the current element?
		require
			a_tag_not_void: a_tag /= Void
		do
			Result :=
				not tags.is_empty and then
				STRING_.same_string (tags.item, a_tag)
		ensure
			definition:
				Result = (not tags.is_empty and then STRING_.same_string (tags.item, a_tag))
		end

	is_tag_started: BOOLEAN is
			-- Is there an unclosed element?
		do
			Result := not tags.is_empty
		end


feature -- Access

	unfinished_xml: STRING is
			-- The `xml' in progress
		do
			Result := my_xml
		end

	as_string: STRING is
			-- The result as plain STRING
		require
			building_finished: not is_tag_started
		do
			Result := my_xml.out
		ensure
			result_not_void: Result /= Void
		end

	as_uc_string: UC_STRING is
			-- The result as Unicode string, i.e. UC_STRING
		require
			building_finished: not is_tag_started
		do
			Result := my_xml
		ensure
			result_not_void: Result /= Void
		end

	xml: STRING is
			-- The result
		obsolete "Use either as_string or as_uc_string"
		do
			Result := as_string
		end


feature -- Influence state

	clear is
			-- Start fresh.
		do
			my_xml.wipe_out
			tags.wipe_out
			clear_attributes
			if is_fragment then
				is_header_written := True
			else
				is_header_written := False
			end
			tag_closed := False
			tag_written := True
		ensure
			no_tags: tags.is_empty
			as_string_empty: as_string.is_empty
			as_uc_string_empty: as_uc_string.is_empty
			header_written: is_fragment = is_header_written
		end


feature -- Commands that expand `xml'

	add_header (encoding: STRING) is
			-- Add the XML header, document is encoded in
			-- `encoding'. Making sure this encoding is followed, is the
			-- responsibility of the client.
		require
			valid_point_for_header: not is_header_written
		do
			extend (xml_header_start)
			extend (encoding)
			extend (xml_header_stop)
			new_line
			is_header_written := True
		ensure
			header_written: is_header_written
		end

	add_header_iso_8859_1_encoding is
			-- Document is iso-8859-1 encoded.
		require
			valid_point_for_header: not is_header_written
		do
			add_header (once_iso_8859_1)
		ensure
			header_written: is_header_written
		end

	add_header_utf_8_encoding is
			-- Document is utf8 encoded.
		require
			valid_point_for_header: not is_header_written
		do
			add_header (once_utf8)
		ensure
			header_written: is_header_written
		end

	add_data, puts (data: STRING) is
			-- Write data in the current tag.
			-- Invalid characters like < or > are quoted.
			-- Use `add_raw' if you don't want quoting.
		require
			valid_point_for_data: is_tag_started
			no_invalid_control_characters: not has_invalid_control_characters (data)
		local
			uc: UC_STRING
			s: STRING
		do
			assure_last_tag_written (False)
			if data /= Void and then not data.is_empty then
				-- With the content being an UC_STRING, `extend_with_data'
				-- seems to be much faster when things have to be
				-- quoted. Do some check to call `extend_with_data' only
				-- when it might be faster.
				-- When there are no meta characters, the first method
				-- should be faster.
				uc ?= data
				if uc = Void and then data.count < 256 then
					s := clone (data)
					replace_content_meta_characters (s)
					extend (s)
				else
					extend_with_data (data)
				end
				is_element_with_data := True
			end
		ensure
			data_added:
				(data /= Void and then not data.is_empty) implies
					is_element_with_data
		end

	add_entity (an_entity_name: STRING) is
			-- Write entity `name' as element data.
		require
			entity_name_not_empty:
				an_entity_name /= Void and then
				not an_entity_name.is_empty
		do
			assure_last_tag_written (False)
			extend_character ('&')
			extend (an_entity_name)
			extend_character (';')
			is_element_with_data := True
		ensure
			data_added: is_element_with_data
		end

	add_raw (raw_data: STRING) is
			-- Write `data' straight in the current tag, meta characters
			-- are not quoted, control characters are not checked, etc.
		do
			assure_last_tag_written (False)
			if raw_data /= Void then
				extend (raw_data)
			end
			is_element_with_data := True
			is_header_written := True
		ensure
			header_written: is_header_written
			data_added: is_element_with_data
		end

	add_system_doctype (root_tag, system_id: STRING) is
			-- Add a <!DOCTYPE element.
			-- Only allowed when no tags have been written.
		require
			have_header: is_header_written
			no_tags_written: not is_tag_started
			valid_root_tag: root_tag /= Void and then not root_tag.is_empty
			valid_system_id: system_id /= Void and then not system_id.is_empty
		do
			extend ("<!DOCTYPE ")
			extend (root_tag)
			extend (" SYSTEM %"")
			extend (system_id)
			extend ("%">")
			new_line
		end

	add_tag (tag, data: STRING) is
			-- Shortcut for `add_tag', `add_data' and `stop_tag'.
		require
			have_header: is_header_written
			valid_tag: is_valid_tag_name (tag)
			no_invalid_control_characters: not has_invalid_control_characters (data)
		do
			start_tag (tag)
			add_data (data)
			stop_tag
		end

	add_ns_tag (name_space, tag, data: STRING) is
			-- Shortcut for `add_ns_tag', `add_data' and `stop_tag'.
		require
			have_header: is_header_written
			valid_name_space:
				name_space = Void or else
				name_space.is_empty or else
				is_valid_ncname (name_space)
			valid_tag: is_valid_ncname (tag)
			no_invalid_control_characters: not has_invalid_control_characters (data)
		do
			start_ns_tag (name_space, tag)
			add_data (data)
			stop_tag
		end

	get_attribute (attribute: STRING): STRING is
			-- Get contents of attribute `attribute' for
			-- current tag. `attribute' may include a name space.
			-- Returns Void if attribute doesn't exist
		do
			if exist_attribute (attribute) then
				Result := get_value
			end
		end

	put (a: ANY) is
			-- Write data within the current tag.
		do
			add_data (a.out)
		ensure
			data_added: is_element_with_data
		end

	put_new_line is
			-- Add a new line in the current tag.
		do
			assure_last_tag_written (False)
			new_line
			is_element_with_data := True
		ensure
			data_added: is_element_with_data
		end

	set_attribute (attribute, value: STRING) is
			-- Set an attribute of the current tag.
			-- `attribute' must be name-space less, else use `set_ns_attribute'.
			-- `value' may not contain an entity reference.
			-- As the attribute is not immediately written, make sure
			-- `attribute' and `value' do not change (ie are cloned or
			-- immutable).
		require
			have_current_tag: is_tag_started
			valid_attribute: is_valid_attribute_name (attribute)
			attribute_has_no_colon: not attribute.has (':')
			no_invalid_control_characters: not has_invalid_control_characters (value)
		do
			if value = Void then
				do_set_attribute (Void, attribute, Void)
			else
				do_set_attribute (Void, attribute, make_valid_attribute_value (value))
			end
		end

	set_a_name_space (a_prefix, a_uri: STRING) is
			-- Define a name space.
			-- As the attribute is not immediately written, make sure
			-- `a_prefix' and `a_uri' do not change (ie are cloned or
			-- immutable).
		require
			have_current_tag: is_tag_started
			prefix_not_empty: a_prefix /= Void and then not a_prefix.is_empty
			uri_not_empty: a_uri /= Void and then not a_uri.is_empty
		do
			do_set_attribute (xmlns, a_prefix, a_uri)
		end

	set_default_name_space (uri: STRING) is
			-- Set the default name space.
		require
			have_current_tag: is_tag_started
			uri_not_empty: uri /= Void and then not uri.is_empty
		do
			do_set_attribute (Void, xmlns, uri)
		end

	set_ns_attribute (name_space, attribute, value: STRING) is
			-- Set an attribute of the current tag.  `value' may not
			-- contain an entity reference. `name_space' is the optional
			-- prefix to be used, not the actual URI.
			-- As the attribute is not immediately written, make sure
			-- `name_space', `attribute' and `value' do not change (ie
			-- are cloned or immutable).
		require
			have_current_tag: is_tag_started
			valid_attribute: is_valid_attribute_name (attribute)
		do
			if value = Void then
				do_set_attribute (name_space, attribute, Void)
			else
				do_set_attribute (name_space, attribute, make_valid_attribute_value (value))
			end
		end

	start_ns_tag (name_space, tag: STRING) is
			-- Start a new tag in the given `name_space'. `name_space' is
			-- a prefix only, not the actual URI. If `name_space' is Void
			-- or empty, the tag will not get a prefix.
			-- As the tag is not immediately written, be sure that `tag'
			-- does not change (ie is cloned or immutable) if
			-- `name_space' is Void or empty.
		require
			have_header: is_header_written
			valid_name_space:
				name_space = Void or else
				name_space.is_empty or else
				is_valid_ncname (name_space)
			valid_tag: is_valid_tag_name (tag)
		local
			ns_tag: UC_UTF8_STRING
		do
			assure_last_tag_written (False)
			if name_space = Void or else name_space.is_empty then
				tags.put (tag)
			else
				create ns_tag.make (name_space.count + 1 + tag.count)
				ns_tag.append_string (name_space)
				ns_tag.append_character (':')
				ns_tag.append_string (tag)
				tags.put (ns_tag)
			end
			tag_written := False
		end

	start_tag (tag: STRING) is
			-- Start a new tag.
			-- As the tag is not immediately written, make sure `tag'
			-- does not change (ie is cloned or immutable).
		require
			have_header: is_header_written
			valid_tag: is_valid_tag_name (tag)
		do
			assure_last_tag_written (False)
			tags.put (tag)
			tag_written := False
		ensure
			tag_not_written: not tag_written
		end

	stop_tag is
			-- Stop last started tag.
		require
			tag_is_started: is_tag_started
		do
			assure_last_tag_written (True)
			if not tag_closed then
				if not is_element_with_data and then on_new_line then
					indent
				end
				extend_character ('<')
				extend_character ('/')
				extend (tags.item)
				extend_character ('>')
				new_line_after_closing_tag (tags.item)
			end
			tags.remove
			is_element_with_data := False
		ensure
			tag_written: tag_written
			no_data: not is_element_with_data
		end


feature {NONE} -- The only methods that operate directly on `my_xml'

	extend_character (c: CHARACTER) is
			-- Add `c' to `my_xml'.
		do
			my_xml.append_character (c)
		ensure
			c_added: my_xml.item (my_xml.count) = c
		end

	extend (stuff: STRING) is
			-- Add anything to the current `xml' string, you're on your own here!
			-- All changes to `my_xml' are made from here.
		require
			stuff_not_void: stuff /= Void
		do
			my_xml.append_string (stuff)
		end

	extend_with_data (data: STRING) is
			-- Extend `my_xml' with `data', but quote any meta characters
			-- that occur in `data'.
		require
			data_not_void: data /= Void
		local
			i: INTEGER
			c: INTEGER
		do
			from
				i := 1
			until
				i > data.count
			loop
				c := data.item_code (i)
				inspect c
				when Less_than_code then
					my_xml.append_string (QuoteLt)
				when Ampersand_code then
					my_xml.append_string (QuoteAmp)
				when Greater_than_code then
						-- Replace ]]> by ]]&gt;
					if
						i > 2 and then
						data.item (i - 1) = ']' and then
						data.item (i - 2) = ']'
					then
						my_xml.append_string (QuoteGt)
					else
						my_xml.append_code (c)
					end
				else
					my_xml.append_code (c)
				end
				i := i + 1
			end
		end


feature -- Quote unsafe characters

	replace_content_meta_characters (s: STRING) is
			-- Replace all characters in `s' that have a special meaning in
			-- XML. These characters are '<' and '&' and the sequence "]]>".
			-- This routine is slow when `data' is actually a UC_STRING
			-- and is very large. Moving bytes to the right to insert the
			-- quoting characters takes up a very long time.
		require
			s_not_void: s /= Void
		local
			i: INTEGER
			code: INTEGER
		do
			from
				i := 1
			variant
				2 + (s.count - i)
			until
				i > s.count
			loop
				code := s.item_code (i)
				inspect code
				when Less_than_code then
					s.put ('&', i)
					s.insert_string (PartQuoteLt, i+1)
					i := i + QuoteLt.count
				when Ampersand_code then
					s.insert_string (PartQuoteAmp, i+1)
					i := i + QuoteAmp.count
				when Greater_than_code then
						-- Replace ]]> by ]]&gt;
					if
						i > 2 and then
						s.item_code (i - 1) = Right_bracket_code and then
						s.item_code (i - 2) = Right_bracket_code
					then
						s.put ('&', i)
						s.insert_string (PartQuoteGt, i+1)
						i := i + QuoteGt.count
					else
						i := i + 1
					end
				else
					i := i + 1
				end
			end
		end


feature -- Comments

	add_comment (a_comment: STRING) is
			-- Add a comment.
		do
			start_comment
			if a_comment /= Void then
				-- Oops, no meta characters to quote here??
				extend (a_comment)
			end
			stop_comment
		end


	start_comment is
			-- Write the XML comment start tag.
		do
			assure_last_tag_written (False)
			extend (start_comment_string)
		end

	stop_comment is
			-- Stop a started XML comment.
		do
			extend (stop_comment_string)
		end


feature {NONE} -- Internal xml change

	assure_last_tag_written (may_close_tag: BOOLEAN) is
			-- Starting a tag is a delayed activity. This routine makes
			-- sure the tag is really written.
		do
			tag_closed := False
			if not tag_written then
				if not is_element_with_data and then not on_new_line then
					new_line_before_starting_tag (tags.item)
				end
				if not is_element_with_data and then on_new_line then
					indent
				end
				extend_character ('<')
				extend (tags.item)
				from
					attributes.start
				until
					attributes.after
				loop
					extend_character (' ')
					extend (attributes.item_for_iteration)
					extend_character ('=')
					extend_character ('"')
					values.search (attributes.item_for_iteration)
						check
							value_exists: values.found
						end
					if values.found_item /= Void then
						extend (values.found_item)
					end
					extend_character ('"')
					attributes.forth
				end
				clear_attributes
				if may_close_tag then
					extend (empty_tag_closing_chars) -- i.e. "/>"
					new_line_after_closing_tag (tags.item)
					tag_closed := True
				else
					extend_character ('>')
				end
				tag_written := True
				is_element_with_data := False
			end
		ensure
			tag_is_written: tag_written
			data_written_is_reset: not old tag_written implies not is_element_with_data
		end

	empty_tag_closing_chars: STRING is
			-- Which characters to use for an empty tag?
			-- XHTML applications like an additional space.
		once
			Result := "/>"
		ensure
			empty_tag_closing_chars_not_empty: Result /= Void and then not Result.is_empty
		end

	indent is
			-- Indent to two spaces under parent.
		local
			spaces: STRING
		do
			inspect tags.count
			when 0, 1 then -- no indent
			when 2 then extend (two_spaces)
			when 3 then extend (four_spaces)
			when 4 then extend (six_spaces)
			when 5 then extend (eight_spaces)
			when 6 then extend (ten_spaces)
			else
				spaces := sh.make_spaces (2 * (tags.count - 1 ) )
				extend (spaces)
			end
		end

	make_valid_attribute_value (value: STRING): STRING is
			-- Convert value to something that is ok within an attribute.
			-- `value' should not contain an entity reference, because it
			-- will be quoted.
			-- Returns either `value' or a new string.
		local
			i: INTEGER
		do
			if value /= Void then
				Result := clone (value)
				from
					i := 1
				until
					i > Result.count
				loop
					inspect Result.item (i)
					when '"' then
						Result.put ('&', i)
						Result.insert_string (PartQuoteQuote, i + 1)
						i := i + QuoteQuote.count
					when '<' then
						Result.put ('&', i)
						Result.insert_string (PartQuoteLt, i + 1)
						i := i + QuoteLt.count
					when '&' then
						Result.insert_string (PartQuoteAmp, i + 1)
						i := i + QuoteAmp.count
					else
						i := i + 1
					end
				end
			end
		end

	new_line is
			-- Write a new line now.
			-- Cannot be called indiscriminately, but only when a pending
			-- tag has been written. Use `put_new_line' for a safe variant.
		do
			extend_character ('%N')
		ensure
			cursor_on_new_line: on_new_line
		end

	new_line_after_closing_tag (a_tag: STRING) is
			-- Outputs a new line, called when `a_tag' is closed
			-- can be overridden to start a new line only occasionally
			-- For XHTML documents a new line is treated as a single
			-- space, so it can influence layout.
		require
			valid_tag: a_tag /= Void and then not a_tag.is_empty
		do
			new_line
		end

	new_line_before_starting_tag (a_tag: STRING) is
			-- Outputs a new line, called when `a_tag' is about to begin.
		require
			valid_tag: a_tag /= Void and then not a_tag.is_empty
		do
			new_line
		end


feature {NONE} -- Stack of tags

	tags: DS_STACK [STRING]
			-- Tags that have been started.


feature {NONE} -- Tag attributes

	attributes: DS_LINKED_LIST [STRING]
			-- Order of attributes.

	values: DS_HASH_TABLE [STRING, STRING]
			-- Attribute values.

	clear_attributes is
		do
			attributes.wipe_out
			values.wipe_out
		end

	do_set_attribute (a_name_space, a_attribute, a_value: STRING) is
			-- Raw set attribute for current tag including optional name space.
			-- Assumes all checks have been done, so `a_value' is
			-- properly quoted for example.
		require
			have_attribute:
				(a_name_space = Void implies
					(a_attribute /= Void and then not a_attribute.is_empty)) and
				(a_attribute = Void implies
					(a_name_space /= Void and then not a_name_space.is_empty))
			attribute_has_no_colon:
				a_attribute /= Void implies not a_attribute.has (':')
		local
			attribute: STRING
		do
			if a_name_space = Void or else a_name_space.is_empty then
				--attribute := to_uc_string (a_attribute)
				attribute := a_attribute
			else
				create {UC_UTF8_STRING} attribute.make (a_name_space.count + 1 + a_attribute.count)
				attribute.append_string (a_name_space)
				attribute.append_character (':')
				attribute.append_string (a_attribute)
			end
			values.search (attribute)
			if values.found then
				values.replace_found_item (a_value)
			else
				attributes.put_last (attribute)
				values.put (a_value, attribute)
			end
		ensure
			attribute_found:
				(a_name_space = Void or else a_name_space.is_empty implies attributes.has (a_attribute)) or else
				(a_name_space /= Void implies attributes.has (a_name_space + ":" + a_attribute))
			value_found:
				(a_name_space = Void or else a_name_space.is_empty implies values.has (a_attribute)) or else
				(a_name_space /= Void implies values.has (a_name_space + ":" + a_attribute))
		end

	exist_attribute (attribute: STRING): BOOLEAN is
			-- Return True if `attribute' exists. `attribute' may contain
			-- a name space. If returns True, `get_value' may be called.
		require
			attribute_not_empty: attribute /= Void and then not attribute.is_empty
		do
			values.search (attribute)
			Result := values.found
		ensure
			found_set: Result implies values.found
		end

	get_value: STRING is
			-- Return the contents of the last searched for attribute
			-- value. A value can be Void.
		require
			value_found: values.found
		do
			Result := values.found_item
		end


feature {NONE} -- Internal state

	my_xml: UC_STRING
			-- The string builded as we go.

	on_new_line: BOOLEAN is
			-- Is cursor on a new line?
		local
			c: INTEGER
		do
			c := my_xml.count
			Result :=
				(c = 0) or else
				(my_xml.item (c) = '%N')
		ensure
			on_new_line:
				Result implies
					my_xml.is_empty or else
					my_xml.item (my_xml.count) = '%N'
		end

	tag_closed: BOOLEAN
			-- Is tag closed?

	tag_written: BOOLEAN
			-- Is the currently started tag written to the stream?
			-- We employ lazy writing so we can output <a/> when there is
			-- no data for example.


feature {NONE} -- Other once strings

	start_comment_string: STRING is "<!--"
	stop_comment_string: STRING is "-->"

	xmlns: STRING is "xmlns"

	xml_header_start: STRING is "<?xml version=%"1.0%" encoding=%""
	xml_header_stop: STRING is "%" ?>"

	once_iso_8859_1: STRING is "ISO-8859-1"
	once_utf8: STRING is "UTF-8"

	two_spaces: UC_UTF8_STRING is
		once
			create Result.make_filled (' ', 2)
		end

	four_spaces: UC_UTF8_STRING is
		once
			create Result.make_filled (' ', 4)
		end

	six_spaces: UC_UTF8_STRING is
		once
			create Result.make_filled (' ', 6)
		end

	eight_spaces: UC_UTF8_STRING is
		once
			create Result.make_filled (' ', 8)
		end

	ten_spaces: UC_UTF8_STRING is
		once
			create Result.make_filled (' ', 10)
		end


feature {NONE} -- Helper routines

	to_uc_string (s: STRING): UC_STRING is
			-- Return `s' as an UC_STRING, needed for proper
			-- hashing.
		obsolete "Not needed anymore, because we use the proper equality tester."
		require
			s_not_empty: s /= Void and then not s.is_empty
		do
			Result ?= s
			if Result = Void then
				create Result.make_from_string (s)
			end
		ensure
			not_void: Result /= Void
			same_string: STRING_.same_string (s, Result)
		end


invariant

	my_xml_not_void: my_xml /= Void
	same_size: attributes.count = values.count
	has_tag_stack: tags /= Void
	comparing_references_is_not_good_enough: tags.equality_tester /= Void
	fragment_has_no_header: is_fragment implies is_header_written
	values_not_void: values /= Void
	attributes_not_void: attributes /= Void
	every_attribute_has_a_value: attributes.count = values.count

end
